from timeit import default_timer as timer
from bson import json_util
import logging

LOGGER = logging.getLogger(__name__)


def db_log_msg(index, req, message, level=None):
    username = req.get_header("username", default="anonymous")
    if level is None:
        LOGGER.info(f"{username} - {message}")
    index._storage.dbLogEvent(message, username=username)
    return


def getMatchingParams(req_params):
    parameters = {}
    for key, value in req_params.items():
        try:
            if key == "pichash_size":
                pichash_size = int(value)
                pichash_size = max(0, pichash_size)
                parameters["pichash_size"] = pichash_size
                # self.index.updatePicHashSize(pichash_size)
            if key == "minhash_score":
                minhash_score = int(value)
                minhash_score = max(0, min(100, minhash_score))
                parameters["minhash_threshold"] = minhash_score
            if key == "force_recalculation":
                if value.lower() == "true":
                    parameters["force_recalculation"] = True
            if key == "band_matches_required":
                band_matches_required = int(value)
                band_matches_required = max(0, band_matches_required)
                parameters["band_matches_required"] = band_matches_required
        except:
            LOGGER.warning(f"Failed to handle request parameter: {key}: {value}")
    return parameters


def jsonify(content, debug_print=False):
    if debug_print:
        print(content)
        print(json_util.dumps(content).encode("utf-8"))
    return json_util.dumps(content).encode("utf-8")


def timing(func):
    def wrapper(*args, **kwargs):
        start = timer()
        func(*args, **kwargs)
        end = timer()
        LOGGER.info("  *** this took: %s sec" % (end - start))

    return wrapper



from dataclasses import dataclass
import datetime
from typing import Dict, TYPE_CHECKING

from mcrit.libs.utility import encode_two_complement, decode_two_complement

if TYPE_CHECKING: # pragma: no cover
    from smda.common.SmdaReport import SmdaReport

#@dataclass
class SampleEntry(object):

    family_id: int
    sample_id: int
    architecture: str
    base_addr: int
    binary_size: int
    binweight: float
    bitness: int
    component: str
    family: str
    version: str
    is_library: bool
    filename: str
    sha256: str
    smda_version: str
    summary: None
    statistics: Dict[str, int]
    timestamp: datetime.datetime

    # TODO -> rename to fromSmdaReport
    def __init__(self, smda_report: "SmdaReport", sample_id=-1, family_id=0):
        self.sample_id = sample_id
        self.family_id = family_id
        if smda_report:
            self.architecture = smda_report.architecture
            self.base_addr = smda_report.base_addr
            self.binary_size = smda_report.binary_size
            self.binweight = smda_report.binweight
            self.bitness = smda_report.bitness
            self.component = smda_report.component
            self.family = smda_report.family
            self.filename = smda_report.filename
            self.is_library = smda_report.is_library
            self.sha256 = smda_report.sha256
            self.smda_version = smda_report.smda_version
            self.statistics = smda_report.statistics.toDict()
            self.timestamp = smda_report.timestamp
            self.version = smda_report.version

    def getShortSha256(self, prefix=8, border=0):
        if border > 0:
            return self.sha256[:border] + "..." + self.sha256[-border:]
        elif prefix > 0:
            return self.sha256[:prefix]
        return self.sha256

    def getShortFilename(self, size_visible=20):
        if len(self.filename) > 2 * size_visible:
            return self.filename[:size_visible] + "..." + self.filename[-size_visible:]
        return self.filename

    def toDict(self):
        sample_entry = {
            "architecture": self.architecture,
            "base_addr": encode_two_complement(self.base_addr),
            "binary_size": self.binary_size,
            "binweight": self.binweight,
            "bitness": self.bitness,
            "component": self.component,
            "family_id": self.family_id,
            "family": self.family,
            "filename": self.filename,
            "is_library": self.is_library,
            "sample_id": self.sample_id,
            "sha256": self.sha256,
            "smda_version": self.smda_version,
            "statistics": self.statistics,
            "timestamp": self.timestamp.strftime("%Y-%m-%dT%H-%M-%S"),
            "version": self.version,
        }
        return sample_entry

    @classmethod
    def fromDict(cls, entry_dict):
        sample_entry = cls(None) #type: ignore
        sample_entry.family_id = entry_dict["family_id"]
        sample_entry.sample_id = entry_dict["sample_id"]
        sample_entry.architecture = entry_dict["architecture"]
        sample_entry.base_addr = decode_two_complement(entry_dict["base_addr"])
        sample_entry.binary_size = entry_dict["binary_size"]
        sample_entry.binweight = entry_dict["binweight"]
        sample_entry.bitness = entry_dict["bitness"]
        sample_entry.component = entry_dict["component"]
        sample_entry.family = entry_dict["family"]
        sample_entry.version = entry_dict["version"]
        sample_entry.is_library = entry_dict["is_library"]
        sample_entry.filename = entry_dict["filename"]
        sample_entry.sha256 = entry_dict["sha256"]
        sample_entry.smda_version = entry_dict["smda_version"]
        sample_entry.statistics = entry_dict["statistics"]
        sample_entry.timestamp = datetime.datetime.strptime(entry_dict["timestamp"], "%Y-%m-%dT%H-%M-%S")
        return sample_entry

    @classmethod
    def fromAlchemySample(cls, sample):
        sample_entry = cls(None) #type:ignore
        sample_entry.sample_id = sample.id
        sample_entry.family_id = sample.family.id
        sample_entry.family = sample.family.name
        sample_entry.architecture = sample.architecture.name
        sample_entry.base_addr = sample.base_addr
        sample_entry.binary_size = sample.binary_size
        sample_entry.component = sample.component
        sample_entry.binweight = sample.binweight
        sample_entry.bitness = sample.bitness.name
        sample_entry.version = sample.version
        sample_entry.is_library = sample.is_library
        sample_entry.filename = sample.filename
        sample_entry.sha256 = sample.sha256
        sample_entry.smda_version = sample.smda_version
        sample_entry.statistics = sample.statistics
        sample_entry.timestamp = sample.timestamp
        return sample_entry

    def __str__(self):
        return "Sample {} ({}, {} bit) - {} ({}): ".format(
            self.sample_id, self.architecture, self.bitness, self.filename, self.family
        )

    def __hash__(self):
        """Override the default hash behavior"""
        return hash(f"{self.sample_id}_{self.sha256}")



#!/usr/bin/env python3
import re
import json
import time
import logging
from datetime import datetime, timedelta
from typing import Dict, List

from smda.common.SmdaReport import SmdaReport

from mcrit.config.McritConfig import McritConfig
from mcrit.config.MinHashConfig import MinHashConfig
from mcrit.config.ShinglerConfig import ShinglerConfig
from mcrit.config.StorageConfig import StorageConfig
from mcrit.index.SearchCursor import MinimalSearchCursor, FullSearchCursor
from mcrit.index.SearchQueryParser import SearchQueryParser
from mcrit.libs.utility import compress_encode, decompress_decode
from mcrit.queue.QueueFactory import QueueFactory
from mcrit.queue.QueueRemoteCalls import QueueRemoteCaller, NoProgressReporter
from mcrit.storage.FamilyEntry import FamilyEntry
from mcrit.storage.FunctionEntry import FunctionEntry
from mcrit.storage.SampleEntry import SampleEntry
from mcrit.storage.MatchedFunctionEntry import MatchedFunctionEntry
from mcrit.storage.StorageFactory import StorageFactory
from mcrit.minhash.MinHash import MinHash
from mcrit.minhash.MinHasher import MinHasher
from mcrit.matchers.MatcherQueryFunction import MatcherQueryFunction
import mcrit.matchers.MatcherInterface as MatcherInterface

from mcrit.Worker import Worker

logging.basicConfig(level=logging.INFO)
LOGGER = logging.getLogger(__name__)


#### Helper methods for nested sort_by in search

def _get_field_once(object, field):
    try:
        return getattr(object, field)
    except Exception:
        pass
    try:
        return object[field]
    except Exception:
        pass

def _get_field(object, field):
    dot_index = field.find(".")
    if dot_index >= 0:
        first_field = field[:dot_index]
        remaining_fields = field[dot_index+1:]
        return _get_field(
            _get_field_once(object, first_field),
            remaining_fields
        )
    else:
        return _get_field_once(object, field)


class MinHashIndex(QueueRemoteCaller(Worker)):
    def __init__(self, config=None, base_url=None):
        if config is None:
            config = McritConfig()
        self.config = config
        self._storage_config = config.STORAGE_CONFIG
        self._minhash_config = config.MINHASH_CONFIG
        self._shingler_config = config.SHINGLER_CONFIG
        self._queue_config = config.QUEUE_CONFIG
        self.minhasher = MinHasher(self._minhash_config, self._shingler_config)
        self._storage = StorageFactory.getStorage(config)
        self._cleanup_delta = timedelta(seconds=int(self._storage_config.STORAGE_MONGODB_CLEANUP_DELTA))
        # config.QUEUE_CONFIG.QUEUE_METHOD = QueueFactory.QUEUE_METHOD_FAKE
        queue = QueueFactory().getQueue(config, storage=self._storage, consumer_id="index")
        self.search_query_parser = SearchQueryParser()
        super().__init__(queue)

    def _getLastCleanup(self):
        try:
            timestamp = self._storage.getDbCleanupTimestamp()
            if not timestamp:
                timestamp = self._storage.updateDbCleanupTimestamp()
            return timestamp
        except Exception as e:
            LOGGER.error(f"Failed getting last cleanup: {e}, this should no automatically fix itself with the next request.")
    
    def _cleanupCallback(self):
        if not self._storage_config.STORAGE_MONGODB_ENABLE_CLEANUP:
            return
        last_timestamp = self._getLastCleanup()
        if last_timestamp:
            now = datetime.now()
            if last_timestamp + self._cleanup_delta < now:
                LOGGER.info("Scheduling a cleanup for query samples.")
                self.doDbCleanup(force_recalculation=True)
                self._storage.updateDbCleanupTimestamp()
        else:
            LOGGER.info("Couldn't determine last db cleanup time, this should no automatically fix itself with the next request.")
            # Should we do cleanup anyway in such cases?

    def _indexCallback(self):
        """ called whenever other functionality in MinHashIndex is used, intended for scheduling maintainance jobs etc. """
        self._cleanupCallback()
    
    #### STORAGE IO ####
    def getStorage(self):
        """Get an interface to the storage"""
        # use this as a callback in all other functions to trigger functionality
        self._indexCallback()
        return self._storage

    def getStorageData(self):
        """Warning: This is intended for local debugging runs - storage may become huge"""
        storage = self.getStorage()
        results = {
            "config": {
                "minhash_config": self._minhash_config.toDict(),
                "shingler_config": self._shingler_config.toDict(),
                "storage_config": self._storage_config.toDict(),
                "queue_config": self._queue_config.toDict(),
            },
            "stats": storage.getStats(),
            "storage": storage.getContent(),
        }
        return results

    def setStorageData(self, storage_data):
        storage = self.getStorage()
        self._minhash_config = MinHashConfig.fromDict(storage_data["config"]["minhash_config"])
        self._shingler_config = ShinglerConfig.fromDict(storage_data["config"]["shingler_config"])
        self._storage_config = StorageConfig.fromDict(storage_data["config"]["storage_config"])
        self._queue_config = StorageConfig.fromDict(storage_data["config"]["queue_config"])
        mcrit_config = McritConfig()
        mcrit_config.MINHASH_CONFIG = self._minhash_config
        mcrit_config.SHINGLER_CONFIG = self._shingler_config
        mcrit_config.STORAGE_CONFIG = self._storage_config
        mcrit_config.QUEUE_CONFIG = self._queue_config
        # reinitialize
        storage = StorageFactory.getStorage(mcrit_config)
        storage.setContent(storage_data["storage"])

    def getExportData(self, sample_ids=None, compress_data=False):
        storage = self.getStorage()
        exported_data = {
            "content": {
                "is_compressed": compress_data,
                "num_families": 0,
                "num_samples": 0,
                "num_functions": 0
            },
            # config hashes need to be identical in order for minhashes to be reusable
            "config": {
                "version": self.config.VERSION,
                "shingler": self._shingler_config.getConfigHash(),
                "minhash": self._minhash_config.getConfigHash(),
            },
            # family mapping is needed for function_entries as they only contain family_ids
            "family_mapping": {},
            # organize by sha256 to simplify avoiding redundancy during import
            "sample_entries": {},
            # organize by sha256 as well, to easily access functions of a sample
            "function_entries": {}
        }
        family_mapping = {}
        exported_sample_entries = {}
        exported_function_entries = {}
        export_size = 0
        # iterate over sample_ids and transform respective data
        for sample_id in storage.getSampleIds():
            if sample_ids and sample_id not in sample_ids:
                continue
            sample_entry = storage.getSampleById(sample_id)
            family_mapping[sample_entry.family_id] = sample_entry.family
            exported_sample_entries[sample_entry.sha256] = sample_entry.toDict()
            function_entries = storage.getFunctionsBySampleId(sample_id)
            functions_dict = {function_entry.function_id: function_entry.toDict() for function_entry in function_entries}
            exported_function_entries[sample_entry.sha256] = functions_dict
            if compress_data:
                compressed_encoded = compress_encode(json.dumps(functions_dict))
                exported_function_entries[sample_entry.sha256] = compressed_encoded
                export_size += len(compressed_encoded)
            else:
                export_size += len(json.dumps(functions_dict))
            exported_data["content"]["num_samples"] += 1
            exported_data["content"]["num_functions"] += len(function_entries)
            if export_size > self._storage_config.STORAGE_MAX_EXPORT_SIZE:
                raise MemoryError("Export running beyond the allocated maximum, aborting operation.")
        exported_data["content"]["num_families"] = len(family_mapping)
        exported_data["family_mapping"] = family_mapping
        exported_data["sample_entries"] = exported_sample_entries
        exported_data["function_entries"] = exported_function_entries
        return exported_data

    def addImportData(self, export_data):
        storage = self.getStorage()
        import_report = {
            "num_samples_imported": 0,
            "num_samples_skipped": 0,
            "num_functions_imported": 0,
            "num_functions_skipped": 0,
            "num_families_imported": 0,
            "num_families_skipped": 0
        }
        # TODO adjust minimum required MCRIT version if we ever extend in a way that objects become incompatible
        # TODO do not compare version as string
        if export_data["config"]["version"] <= "0.0.0":
            LOGGER.error("Cannot import data, version of export is incompatible / too old.")
            return 
        if export_data["config"]["shingler"] != self._shingler_config.getConfigHash():
            LOGGER.error("Cannot import data, shingler configuration hash is incompatible.")
            return 
        if export_data["config"]["minhash"] != self._minhash_config.getConfigHash():
            LOGGER.error("Cannot import data, MinHash configuration hash is incompatible.")
            return 
        is_compressed = export_data["content"]["is_compressed"]
        # create a dictionary for pointing family_ids as contained in the export to family_ids as used in this instance
        family_id_remapping = {}
        max_family_id_in_storage = max(storage.getFamilyIds())
        for exported_family_id, exported_family in export_data["family_mapping"].items():
            exported_family_id = int(exported_family_id)
            remapped_family_id = storage.addFamily(exported_family)
            if remapped_family_id > max_family_id_in_storage:
                import_report["num_families_imported"] += 1
                max_family_id_in_storage = remapped_family_id
            else:
                import_report["num_families_skipped"] += 1
            family_id_remapping[exported_family_id] = remapped_family_id
        LOGGER.info("Family remapping created: %d families, %d samples.", 
                len(family_id_remapping), 
                len(export_data["sample_entries"]))
        # iterate samples
        index = 0
        function_entries_to_import = []
        for sample_sha256, sample_entry_dict in export_data["sample_entries"].items():
            index += 1
            # check if sample is already present, and skip if yes
            if storage.getSampleBySha256(sample_sha256):
                LOGGER.info(f"Sample with SHA256 {sample_sha256} already present in database, skipping...")
                import_report["num_samples_skipped"] += 1
                import_report["num_functions_skipped"] += sample_entry_dict["statistics"]["num_functions"]
                continue
            import_report["num_samples_imported"] += 1
            sample_entry = SampleEntry.fromDict(sample_entry_dict)
            # adjust family_id in sample_entry using our remapping
            sample_entry.family_id = family_id_remapping[sample_entry.family_id]
            # add sample_entry to storage and receive new sample_id
            remapped_sample_entry = storage.importSampleEntry(sample_entry)
            if sample_sha256 in export_data["function_entries"]:
                function_entries = export_data["function_entries"][sample_sha256]
                # decompress function_entries if necessary
                if is_compressed:
                    function_entries = json.loads(decompress_decode(function_entries))
                # iterate functions and add them, adjusting family_id and sample_id
                for old_function_id, function_entry_dict in function_entries.items():
                    function_entry = FunctionEntry.fromDict(function_entry_dict)
                    function_entry.sample_id = remapped_sample_entry.sample_id
                    function_entry.family_id = remapped_sample_entry.family_id
                    function_entries_to_import.append(function_entry)
                    import_report["num_functions_imported"] += 1
            LOGGER.info(f"Sample %d with SHA256 %s added...", index, sample_sha256)
        
        adjusted_function_entries = storage.importFunctionEntries(function_entries_to_import)
        LOGGER.info(f"{len(adjusted_function_entries)} functions added")
        # we can only start indexing once our imported function_entries had their function_id adjusted to our DB
        minhashes_to_import = []
        for function_entry in adjusted_function_entries:
            # delaying minhash insertion speeds up the procedure dramatically
            if function_entry.minhash:
                minhash = MinHash(function_entry.function_id, function_entry.minhash, minhash_bits=self._minhash_config.MINHASH_SIGNATURE_BITS)
                minhashes_to_import.append(minhash)
        # ensure that their minhashes / pichashes are added to the respective indices
        storage.addMinHashes(minhashes_to_import)
        LOGGER.info(f"{len(minhashes_to_import)} Minhashes added")
        return import_report

    def respawn(self):
        """Get an interface to the storage"""
        storage = self.getStorage()
        self.queue.clear()
        return storage.clearStorage()

    #### REDIRECTED TO WORKER ####
    """
    def updateMinHashes(self, function_ids):
    def rebuildIndex(self):
    def recalculatePicHashes(self):
    def recalculateMinHashes(self):
    def getMatchesForReport(self, report):
    def getMatchesForSmdaReport(self, report_json, minhash_threshold=None):
    def getMatchesForMappedBinary(self, binary, base_address, minhash_threshold=None):
    def getMatchesForUnmappedBinary(self, binary, minhash_threshold=None):
    def getMatchesForSample(self, sample_id, minhash_threshold=None):
    def getMatchesForSampleVs(self, sample_id, other_sample_id, minhash_threshold=None):
    def getAggregatedMatchesForSample(self, sample_id, minhash_threshold=None):
    def getUniqueBlocks(self, sample_ids):
    def addBinarySample(self, binary, is_dump, bitness, base_address,):
    # these need to be jobs to ensure database consistency
    def deleteSample(self, sample_id):
    def deleteFamily(self, family_id, keep_samples=False):
    def modifyFamily(self, family_id, update_information):
    def modifySample(self, sample_id, update_information):
    """

    #### NOT REDIRECTED ####
    def addReport(self, smda_report, calculate_hashes=True, calculate_matches=False, username=None):
        storage = self.getStorage()
        sample_entry = storage.getSampleBySha256(smda_report.sha256)
        if sample_entry:
            self._storage.updateFunctionLabels(smda_report, username)
            return {"existed": True, "sample_info": sample_entry.toDict()}
        sample_entry = storage.addSmdaReport(smda_report)
        if not sample_entry:
            return None
        LOGGER.info("Added %s", sample_entry)
        # ensure that original function_names are also added as labels
        storage.updateFunctionLabels(smda_report, username)
        function_entries = storage.getFunctionsBySampleId(sample_entry.sample_id)
        LOGGER.info("Added %d function entries.", len(function_entries))
        job_id = None
        if calculate_hashes:
            job_id = self.updateMinHashesForSample(sample_entry.sample_id)
        return {"existed": False, "sample_info": sample_entry.toDict(), "job_id": job_id}

    def addReportJson(self, report_json, calculate_hashes=True, calculate_matches=False, username=None):
        storage = self.getStorage()
        report = SmdaReport.fromDict(report_json)
        return self.addReport(report, calculate_hashes=calculate_hashes, calculate_matches=calculate_matches, username=username)

    def addReportFile(self, report_filepath, calculate_hashes=True, calculate_matches=False):
        storage = self.getStorage()
        with open(report_filepath, "r") as fin:
            report_json = json.load(fin)
        report = SmdaReport.fromDict(report_json)
        return self.addReport(report, calculate_hashes=calculate_hashes, calculate_matches=calculate_matches)
    
    def getMatchesCross(self, sample_ids:List[int], force_recalculation=False, **params):
        storage = self.getStorage()
        sample_to_job_id = {}
        for id in sample_ids:
            job_id = self.getMatchesForSample(id, force_recalculation=force_recalculation, **params)
            sample_to_job_id[id] = job_id
        return self.combineMatchesToCross(sample_to_job_id, await_jobs=[*sample_to_job_id.values()], force_recalculation=force_recalculation)

    def getMatchesForSmdaFunction(self, 
            smda_report_with_function:SmdaReport,
            minhash_threshold=None, 
            pichash_size=None, 
            band_matches_required=None, 
            exclude_self_matches=False
        ):
        storage = self.getStorage()
        # convert function to FunctionEntry
        smda_report = SmdaReport.fromDict(smda_report_with_function)
        function_offset = None
        if len(smda_report.xcfg) != 1:
            raise ValueError("SmdaReport has to contain exactly one function.")
        function_offset = int([k for k in smda_report.xcfg.keys()][0])
        matcher = MatcherQueryFunction(self,
            minhash_threshold=None, 
            pichash_size=None, 
            band_matches_required=band_matches_required, 
            exclude_self_matches=False
        )
        # run Matcher for a single function
        match_report = matcher.getMatchesForSmdaFunction(smda_report)
        function_identifier = f"{smda_report.sha256[:8]}@0x{function_offset:x}"
        match_report["info"]["job"]["parameters"] = f"getMatchesForSmdaFunction({function_identifier})"
        # return completed MatchingReport
        return match_report

    def getMatchesFunctionVs(self, function_id_a:int, function_id_b:int):
        storage = self.getStorage()
        function_entry_a = storage.getFunctionById(function_id_a, with_xcfg=True)
        function_entry_b = storage.getFunctionById(function_id_b, with_xcfg=True)
        if function_entry_a is None or function_entry_b is None:
            return
        sample_entry_a = storage.getSampleById(function_entry_a.sample_id)
        sample_entry_b = storage.getSampleById(function_entry_b.sample_id)
        minhash_a = function_entry_a.getMinHash(minhash_bits=self._minhash_config.MINHASH_SIGNATURE_BITS)
        minhash_b = function_entry_b.getMinHash(minhash_bits=self._minhash_config.MINHASH_SIGNATURE_BITS)

        score = None
        if minhash_a.minhash and minhash_b.minhash:
            score = MinHash.calculateMinHashScore(
                    minhash_a.minhash, minhash_b.minhash, minhash_bits=self._minhash_config.MINHASH_SIGNATURE_BITS
                )
        match_flags = 0
        match_flags += MatcherInterface.IS_MINHASH_FLAG if score is not None and score >= self._minhash_config.MINHASH_MATCHING_THRESHOLD else 0
        match_flags += MatcherInterface.IS_PICHASH_FLAG if function_entry_a.pichash == function_entry_b.pichash else 0
        match_flags += MatcherInterface.IS_LIBRARY_FLAG if sample_entry_b.is_library else 0
        match_tuple = [
            function_entry_b.family_id, 
            function_entry_b.sample_id, 
            function_entry_b.function_id, 
            score, 
            match_flags]
        result = {
            "function_entry_a": function_entry_a.toDict(),
            "function_entry_b": function_entry_b.toDict(),
            "sample_entry_a": sample_entry_a.toDict(),
            "sample_entry_b": sample_entry_b.toDict(),
            "match_entry": MatchedFunctionEntry(function_id_a, function_entry_a.binweight, function_entry_a.offset, match_tuple).toDict()
        }
        return result

    #### SIMPLE LOOKUPS ####
    def getFamily(self, family_id):
        return self.getStorage().getFamily(family_id)

    def getFunctionsBySampleId(self, sample_id):
        return self.getStorage().getFunctionsBySampleId(sample_id)

    def isFunctionId(self, function_id):
        return self.getStorage().isFunctionId(function_id)

    def isSampleId(self, sample_id):
        return self.getStorage().isSampleId(sample_id)

    def getFunctionById(self, function_id, with_xcfg=False):
        return self.getStorage().getFunctionById(function_id, with_xcfg=with_xcfg)

    def getFunctions(self, start_index, limit):
        return self.getStorage().getFunctions(start_index, limit)

    def getSamplesByFamilyId(self, family_id):
        return self.getStorage().getSamplesByFamilyId(family_id)

    def getSampleById(self, sample_id):
        return self.getStorage().getSampleById(sample_id)

    def getSamples(self, start_index, limit):
        return self.getStorage().getSamples(start_index, limit)

    def getFamilies(self) -> Dict[int, FamilyEntry]:
        storage = self.getStorage()
        family_overview = {}
        for family_id in storage.getFamilyIds():
            family_overview[family_id] = storage.getFamily(family_id)
        return family_overview

    def getFunctionGraph(self, function_id):
        # NOTE: Does not work at the moment, as xcfg might be deleted.
        raise NotImplementedError
        function_entry = self._storage.getFunctionById(function_id)
        function_info = {"cfg": function_entry.xcfg}
        return function_info

    def getAllSampleInfos(self):
        infos = []
        for sample_id in sorted(self.getStorage().getSampleIds()):
            infos.append(self.getSampleById(sample_id))
        return infos

    def getStatus(self, with_pichash=True):
        storage_stats = self.getStorage().getStats(with_pichash=with_pichash)
        status = {
            "status": {
                "db_state": storage_stats["db_state"],
                "db_timestamp": storage_stats["db_timestamp"],
                "storage_type": self._storage_config.STORAGE_METHOD,
                "num_bands": storage_stats["num_bands"],
                "num_samples": storage_stats["num_samples"],
                "num_families": storage_stats["num_families"],
                "num_functions": storage_stats["num_functions"],
                "num_pichashes": storage_stats["num_pichashes"],
            }
        }
        return status

    def getVersion(self):
        storage = self.getStorage()
        return {"version": self.config.VERSION}

    def getMatchesForPicHash(self, pichash):
        return self.getStorage().getMatchesForPicHash(pichash)
        
    def getMatchesForPicBlockHash(self, picblockhash):
        return self.getStorage().getMatchesForPicBlockHash(picblockhash)

    def getSampleBySha256(self, sample_sha256): 
        return self.getStorage().getSampleBySha256(sample_sha256)

    #### SEARCH ####

    # When performing an initial search, the cursor should be set to None.
    # Search results are of the following form:
    # {
    #     "search_results": {
    #         id1: found_entry1,
    #         id2: found_entry2,
    #         ...
    #     },
    #     "cursor": {
    #         "forward": forward cursor,
    #         "backward": backward cursor,
    #     } 
    # }
    # To get further results, perform a search using the forward cursor.
    # To get back to the previous search results, use the backward cursor.
    # If no further or previous results are available, the forward or backward cursor will be None.
    #
    # IMPORTANT: A cursor shall only be used in combination with the same
    # search_term and sort_by_list that were used when the cursor was returned from mcrit.
    # If those parameters are altered, mcrit's behavior is undefined.

    def _getSearchResultTemplate(self, search_function, search_term, sort_by_list, cursor_str, limit):
        assert isinstance(search_term, str)

        sort_fields = [sort_info[0] for sort_info in sort_by_list]

        is_backward_search = False
        cursor = MinimalSearchCursor.fromStr(cursor_str)

        full_cursor = FullSearchCursor(cursor, sort_by_list)
        is_backward_search = not full_cursor.is_forward_search

        parsed_search = self.search_query_parser.parse(search_term)
        search_results_objects = search_function(parsed_search, cursor=full_cursor, max_num_results=limit+1)

        # Find last last_element_key, which is used for the forward cursor
        search_results_keys = list(search_results_objects.keys())
        if len(search_results_objects) >= limit + 1:
            search_results_objects.pop(search_results_keys[-1])
            search_results_keys.pop()
            last_element_key = search_results_keys[-1]
        else:
            last_element_key = None
        
        forward_cursor_str = None
        if last_element_key:
            last_result = search_results_objects[last_element_key]
            forward_cursor = MinimalSearchCursor()
            forward_cursor.is_forward_search = True ^ is_backward_search # switch for backward search, because of swap
            forward_cursor.record_values = [_get_field(last_result, field) for field in sort_fields]
            forward_cursor_str = forward_cursor.toStr()

        backward_cursor_str = None
        if len(search_results_objects) > 0 and cursor is not None:
            first_result = search_results_objects[search_results_keys[0]]
            backward_cursor = MinimalSearchCursor()
            backward_cursor.is_forward_search = False ^ is_backward_search # switch for backward search, because of swap
            backward_cursor.record_values = [_get_field(first_result, field) for field in sort_fields]
            backward_cursor_str = backward_cursor.toStr()

        if is_backward_search:
            # reverse order
            backward_cursor_str, forward_cursor_str = forward_cursor_str, backward_cursor_str
            search_results = {k: v.toDict() for k, v in reversed(search_results_objects.items())}
        else:
            search_results = {k: v.toDict() for k, v in search_results_objects.items()}

        return {
            "search_results": search_results,
            "cursor": {
                "forward": forward_cursor_str,
                "backward": backward_cursor_str,
            }
        }

    def _get_sort_data(self, standard_sort, sort_by, is_ascending):
        if sort_by == standard_sort or sort_by is None:
            sort_by_list = [(standard_sort, is_ascending),]
        else:
            sort_by_list = [
                (sort_by, is_ascending),
                (standard_sort, True),
            ]
        return sort_by_list 

    def getFamilySearchResults(self, search_term, sort_by="family_id", is_ascending=True, cursor=None, limit=100):
        storage = self.getStorage()
        term_as_int = None
        id_match = None
        try:
            if search_term.startswith("0x"):
                term_as_int = int(search_term, 16)
            else:
                term_as_int = int(search_term)
            if term_as_int <= 0xFFFFFFFF:
                if storage.isFamilyId(term_as_int):
                    id_match = storage.getFamily(term_as_int).toDict()
            else:
                LOGGER.warning("Can only handle family/sample/function IDs up to 0xFFFFFFFF.")
        except Exception:
            pass

        assert sort_by in (
            None,
            "family_id",
            "family_name",
            "num_samples",
            "num_library_samples",
            "num_functions",
        )

        sort_data = self._get_sort_data("family_id", sort_by, is_ascending)
        result = self._getSearchResultTemplate(
            storage.findFamilyByString,
            search_term,
            sort_data,
            cursor,
            limit,
        )
        result["id_match"] = id_match
        return result

    def getFunctionSearchResults(self, search_term, sort_by="function_id", is_ascending=True, cursor=None, limit=100):
        storage = self.getStorage()
        term_as_int = None
        id_match = None
        try:
            if search_term.startswith("0x"):
                term_as_int = int(search_term, 16)
            else:
                term_as_int = int(search_term)
            if term_as_int <= 0xFFFFFFFF:
                if storage.isFunctionId(term_as_int):
                    id_match = storage.getFunctionById(term_as_int).toDict()
            else:
                LOGGER.warning("Can only handle family/sample/function IDs up to 0xFFFFFFFF.")
        except Exception:
            pass

        assert sort_by in (
            None,
            "function_id",
            "sample_id",
            "family_id",
            "pichash",
            "function_name",
            "offset",
            "num_instructions",
            "num_blocks",
        )
        
        sort_data = self._get_sort_data("function_id", sort_by, is_ascending)
        result = self._getSearchResultTemplate(
            storage.findFunctionByString,
            search_term,
            sort_data,
            cursor,
            limit,
        )
        result["id_match"] = id_match
        return result

    def getSampleSearchResults(self, search_term, sort_by="sample_id", is_ascending=True, cursor=None, limit=100):
        storage = self.getStorage()
        term_as_int = None
        id_match = None
        try:
            if search_term.startswith("0x"):
                term_as_int = int(search_term, 16)
            else:
                term_as_int = int(search_term)
            if term_as_int <= 0xFFFFFFFF:
                if storage.isSampleId(term_as_int):
                    id_match = storage.getSampleById(term_as_int).toDict()
            else:
                LOGGER.warning("Can only handle family/sample/function IDs up to 0xFFFFFFFF.")
        except Exception:
            pass

        if re.match("^[a-fA-F0-9]{64}$", search_term) is not None:
            sample_entry = storage.getSampleBySha256(search_term)
            sha_match = sample_entry.toDict()
        else:
            sha_match = None

        assert sort_by in (
            None,
            "filename",
            "function_id",
            "sample_id",
            "family_id",
            "family",
            "architecture",
            "base_addr",
            "binary_size",
            "binweight",
            "bitness",
            "component",
            "is_library",
            "sha256",
            "timestamp",
            "version",
            "statistics.num_functions"
        )
        
        sort_data = self._get_sort_data("sample_id", sort_by, is_ascending)
        result = self._getSearchResultTemplate(
            storage.findSampleByString,
            search_term,
            sort_data,
            cursor,
            limit,
        )
        result["id_match"] = id_match
        result["sha_match"] = sha_match
        return result


    ##### CONFIG CHANGES ####
    def updateMinHashThreshold(self, threshold):
        storage = self.getStorage()
        self.config.MINHASH_CONFIG.MINHASH_MATCHING_THRESHOLD = threshold
        self._minhash_config.MINHASH_MATCHING_THRESHOLD = threshold

    def updatePicHashSize(self, size):
        storage = self.getStorage()
        self.config.MINHASH_CONFIG.PICHASH_SIZE = size
        self._minhash_config.PICHASH_SIZE = size

    def updateMinHasherConfig(self, config):
        storage = self.getStorage()
        self._storage_config = config.STORAGE_CONFIG
        self._minhash_config = config.MINHASH_CONFIG
        self._shingler_config = config.SHINGLER_CONFIG



import logging
import os

from .MinHashConfig import MinHashConfig
from .QueueConfig import QueueConfig
from .ShinglerConfig import ShinglerConfig
from .StorageConfig import StorageConfig


class McritConfig(object):

    # NOTE to self: always change this in setup.py as well!
    VERSION = "1.3.11"
    # basic pathing info
    CONFIG_FILE_PATH = str(os.path.abspath(__file__))
    PROJECT_ROOT = str(os.path.abspath(os.sep.join([CONFIG_FILE_PATH, "..", ".."])))

    # Authentication token, which can be optionally used to lock down communication with the API
    AUTH_TOKEN = ""

    ### global logging-config setup
    # Only do basicConfig if no handlers have been configured
    LOG_PATH = "./"
    LOG_LEVEL = logging.INFO
    LOG_FORMAT = "%(asctime)-15s: %(name)-32s - %(message)s"

    # initialize context-based configurations
    MINHASH_CONFIG = MinHashConfig()
    SHINGLER_CONFIG = ShinglerConfig()
    STORAGE_CONFIG = StorageConfig()
    QUEUE_CONFIG = QueueConfig()

    def __init__(self, log_level=logging.INFO):
        if len(logging._handlerList) == 0:
            logging.basicConfig(level=log_level, format=self.LOG_FORMAT)



from typing import Dict, Set, TYPE_CHECKING, Tuple
from mcrit.matchers.MatcherInterface import MatcherInterface, add_duration
from mcrit.storage.FunctionEntry import FunctionEntry
from mcrit.storage.MatchingCache import MatchingCache
from mcrit.storage.SampleEntry import SampleEntry

if TYPE_CHECKING:
    from smda.common.SmdaReport import SmdaReport

class MatcherQuery(MatcherInterface):
    def _additional_setup(self):
        self._sample_entry = None
        self._sample_id = -1
        # initialize with query object
        self._sample_to_lib_info = {-1: False}
        self._sample_id_to_entry = {-1: 0}

    @add_duration
    def getMatchesForSmdaReport(self, smda_report: "SmdaReport"):
        # create temporary objects similar to the stored samples/functions
        self._sample_entry = self._storage.addSmdaReport(smda_report, isQuery=True)
        self._sample_id = self._sample_entry.sample_id
        self._sample_id_to_entry[self._sample_entry.sample_id] = self._sample_entry
        self._sample_info = self._sample_entry.toDict()
        tmp_function_entries_dict = {}
        for function_entry in self._storage.getFunctionsBySampleId(self._sample_entry.sample_id):
            tmp_function_entries_dict[function_entry.function_id] = function_entry

        minhashes = self._worker.calculateMinHashes(tmp_function_entries_dict.values())
        for minhash in minhashes:
            tmp_function_entries_dict[minhash.function_id].minhash = minhash.getMinHash()
            tmp_function_entries_dict[minhash.function_id].minhash_shingle_composition = minhash.getComposition()
        self._function_entries = list(tmp_function_entries_dict.values())

        return self._getMatchesRoutine()

    def _getPicHashMatches(self) -> Dict[int, Set[Tuple[int, int, int]]]:
        pichash_matches = {}
        checked_pichashes = set()
        for function_entry in self._function_entries:
            if function_entry.pichash:
                if function_entry.pichash not in checked_pichashes:
                    checked_pichashes.add(function_entry.pichash)
                    matches = self._storage.getMatchesForPicHash(function_entry.pichash)
                    if len(matches) > 0:
                        pichash_matches[function_entry.pichash] = matches
                if function_entry.pichash in pichash_matches:
                    pichash_matches[function_entry.pichash].add((function_entry.family_id, function_entry.sample_id, function_entry.function_id))
        return pichash_matches

    def _createMatchingCache(self, candidate_groups):
        function_ids_from_storage = set()
        for own_function_id, other_function_ids in candidate_groups.items():
            for function_id in other_function_ids:
                if function_id >= 0:
                    function_ids_from_storage.add(function_id)
        cache = self._storage.createMatchingCache(function_ids_from_storage)
        cache.addFunctionEntriesToCache(self._function_entries)
        return cache



# from mcrit.storage.alchemy.AlchemyStorage import AlchemyStorage
from mcrit.storage.MemoryStorage import MemoryStorage
from mcrit.storage.MongoDbStorage import MongoDbStorage


class StorageFactory:

    STORAGE_METHOD_MEMORY = "memory"
    STORAGE_METHOD_ALCHEMY = "alchemy"
    STORAGE_METHOD_MONGODB = "mongodb"

    @staticmethod
    def getStorage(mcrit_config):
        if mcrit_config.STORAGE_CONFIG.STORAGE_METHOD == StorageFactory.STORAGE_METHOD_MONGODB:
            return MongoDbStorage(mcrit_config)
        # Alchemy not supported right now
        # if storage_config.STORAGE_METHOD == StorageFactory.STORAGE_METHOD_ALCHEMY:
        #     return AlchemyStorage(storage_config)
        return MemoryStorage(mcrit_config)



import hashlib
import logging
import os
from dataclasses import asdict, dataclass, field
from typing import Dict

from mcrit.config.ConfigInterface import ConfigInterface, default_field
from mcrit.minhash.ShingleLoader import ShingleLoader


@dataclass
class ShinglerConfig(ConfigInterface):

    THIS_FILE_PATH: str = str(os.path.abspath(__file__))
    PROJECT_ROOT: str = str(os.path.abspath(os.sep.join([THIS_FILE_PATH, "..", ".."])))

    # Directory in which to search for files matching pattern "*Shingler.py"
    SHINGLER_DIR: str = str(os.path.abspath(os.sep.join([PROJECT_ROOT, "shinglers"])))
    # Application of Shinglers can be influenced by weights, effectively running the same instance multiple times (increasing their "chance")
    SHINGLER_WEIGHT_STRATEGY: ... = ShingleLoader.WEIGHT_STRATEGY_SHINGLER_WEIGHTS
    # the expected range for logbucket matching
    SHINGLER_LOGBUCKETS: int = 100000
    # number of values to put in the range left and right
    SHINGLER_LOGBUCKET_RANGE: int = 1
    # add additional counts for values further to the center
    SHINGLER_LOGBUCKET_CENTERED: bool = True
    # The weights to use for the above described method
    SHINGLERS_WEIGHTS: ... = default_field(
        {"FuzzyStatPairShingler": 1, "EscapedBlockShingler": 3}
    )
    # random seed to be used when initiating XOR values for shingler hash seeds
    SHINGLERS_SEED: int = 0xDEADBEEF
    # will be set automatically when shinglers are loaded (listed here for completeness)
    SHINGLERS_XOR_VALUES: ... = default_field([])

    def getConfigHash(self):
        config_str = ""
        config_str += f"_{self.SHINGLER_WEIGHT_STRATEGY}_{self.SHINGLER_LOGBUCKETS}_{self.SHINGLER_LOGBUCKET_RANGE}_{self.SHINGLER_LOGBUCKET_CENTERED}"
        config_str += f"_{','.join(['%s-%s' % (k, v) for k, v in sorted(self.SHINGLERS_WEIGHTS.items())])}"
        config_str += f"_{self.SHINGLERS_SEED}"
        return hashlib.sha256(config_str.encode('utf-8')).hexdigest()



import base64
import json
from typing import Any, List, Optional

from mcrit.index.SearchQueryTree import AndNode, OrNode, SearchConditionNode


class MinimalSearchCursor:
    """
    Contains the minimum amount of data (when combined with search order info) required to reconstruct a FullSearchCursor.
    """
    record_values: List[Any]
    is_forward_search: bool

    def __init__(self):
        self.record_values = []
        self.is_forward_search = True

    def toDict(self):
        result = self.record_values, 1 if self.is_forward_search else 0
        return result

    @classmethod 
    def fromDict(cls, data):
        result = cls()
        result.record_values, result.is_forward_search = data[0], bool(data[1])
        return result

    def toStr(self):
        return base64.urlsafe_b64encode(json.dumps(self.toDict()).encode()).decode()
        # return json.dumps(self.toDict())

    @classmethod 
    def fromStr(cls, data):
        # return cls.fromDict(json.loads(data))
        if data is not None:
            return cls.fromDict(json.loads(base64.urlsafe_b64decode(data).decode()))


class FullSearchCursor:
    """
    Contains Data to deliver sorted search results after a certain point.
    sort_fields:
        is the list of n fields used for sorting.
        Fields with higher sorting priority come first.
        These fields together should be unique for each db entry.
    sort_directions:
        a list of n booleans corresponding to the n fields.
        True means ascending order, False descending
    record_values: 
        the n concrete field values of a record.
        The SearchCursor references those records within the db
        which come either after or before this record.
        This record itself
    is_forward_search:
        if True, the SearchCursor references records after the specified record.
        if False, the SearchCursor references records befor the specified record.
    is_initial_cursor:
        True if record_values contains no meaningful data and the cursor should not exclude any records 
    """
    sort_fields: List[str]
    sort_directions: List[bool]
    record_values: List[Any]
    is_forward_search: bool
    is_initial_cursor: bool

    def __init__(self, input_cursor: Optional[MinimalSearchCursor], sort_by_list):
        """
        sort_by_list:
            a list of (sort_field, sort_direction) entries    
        """
        self.sort_fields = [sort_info[0] for sort_info in sort_by_list]
        self.sort_directions = [sort_info[1] for sort_info in sort_by_list]
        if input_cursor is not None:
            self.record_values = input_cursor.record_values
            self.is_forward_search = input_cursor.is_forward_search
            self.is_initial_cursor = False
        else:
            self.record_values = []
            self.is_forward_search = True
            self.is_initial_cursor = True

    @property
    def sort_by_list(self):
        return zip(self.sort_fields, self.sort_directions)
    
    def toTree(self):
        def get_operator(i):
            direction = self.sort_directions[i] ^ (not self.is_forward_search)
            return ">" if direction else "<"

        if self.is_initial_cursor:
            return AndNode([])

        # condition has form (a > a0) or (a = a0 and b>b0) or (a=a0 and b=b0 and c>c0)...
        conditions = []
        for inner_condition_length in range(1, len(self.sort_fields)+1):
            inner_condition = []
            for i in range(inner_condition_length):
                if i != inner_condition_length-1:
                    operator = "="
                else:
                    operator = get_operator(i)
                inner_condition.append(SearchConditionNode(self.sort_fields[i], operator, self.record_values[i]))
            conditions.append(AndNode(inner_condition))

        return OrNode(conditions)





class MatchingCache(object):
    """A reduced in-memory view for a selection of FunctionEntry and corresponding SampleEntry objects - implements a subset of StorageInterface"""

    def __init__(self, cache_data):
        self._func_id_to_minhash = cache_data["func_id_to_minhash"]
        self._func_id_to_sample_id = cache_data["func_id_to_sample_id"]
        self._sample_id_to_func_ids = cache_data["sample_id_to_func_ids"]

    def isSampleId(self, sample_id):
        return sample_id in self._sample_id_to_func_ids

    def getMinHashByFunctionId(self, function_id):
        return self._func_id_to_minhash[function_id]

    def getSampleIdByFunctionId(self, function_id):
        return self._func_id_to_sample_id[function_id]

    def getFunctionIdsBySampleId(self, sample_id):
        return self._sample_id_to_func_ids[sample_id]

    def addFunctionEntriesToCache(self, function_entries):
        for function_entry in function_entries:
            self._func_id_to_minhash[function_entry.function_id] = function_entry.minhash
            self._func_id_to_sample_id[function_entry.function_id] = function_entry.sample_id
            sample_id = function_entry.sample_id
            if sample_id not in self._sample_id_to_func_ids:
                self._sample_id_to_func_ids[sample_id] = set()
            self._sample_id_to_func_ids[sample_id].add(function_entry.function_id)



import os
import sys


def runWorker(profiling=False):
    from mcrit.Worker import Worker
    with Worker(profiling=profiling) as worker:
        worker.run()


def runServer(profiling=False, uses_gunicorn=False):
    import platform
    from waitress import serve
    from mcrit.server.wsgi import app
    from mcrit.config.GunicornConfig import GunicornConfig
    try:
        import gunicorn
        from gunicorn.app.base import BaseApplication
    except:
        gunicorn = None

    class gunicornServer(BaseApplication):
        def __init__(self, app):
            self.app = app
            super().__init__()
        
        def load_config(self):
            for key, value in GunicornConfig().toDict().items():
                try:
                    self.cfg.set(key.lower(), value)
                except AttributeError:
                    continue

        def load(self):
            return self.app

    wrapped_app = app
    if profiling:
        print("[!] Running as profiled application.")
        from werkzeug.middleware.profiler import ProfilerMiddleware
        profile_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "profiler")
        os.makedirs(profile_dir, exist_ok=True)
        wrapped_app = ProfilerMiddleware(
            wrapped_app,
            restrictions=[30],
            profile_dir=profile_dir,
            filename_format="{method}-{path}-{time:.0f}-{elapsed:.0f}ms.prof",
        )
    
    platform = platform.system().lower()
    if platform == "linux" and gunicorn is not None and (GunicornConfig().USE_GUNICORN or uses_gunicorn):
        print("[!] Detected linux platform and gunicorn availability. Using gunicorn deployment.")
        gunicornServer(wrapped_app).run()
        sys.exit()
    elif platform == "windows":
        print("[!] Detected windows platform. Using waitress deployment.")
    else:
        print("[!] Could not determine platform, gunicorn not available or activated. Defaulting to waitress deployment.")
    # TODO consider allowing an argument to pass an configuration for initial setup of the instance
    serve(wrapped_app, listen="*:8000")


def runClient():
    from mcrit.client.McritConsole import McritConsole
    # we will re-parse arguments with more detail in McritConsole to keep this file concise
    console_client = McritConsole()
    console_client.run()


# do not use argparse for processing here to allow using argparse help for the more intricate things in McritConsole
if len(sys.argv) >= 2:
    is_profiling = False
    uses_gunicorn = False
    if len(sys.argv) >= 3 and "--profile" in sys.argv[2:]:
        is_profiling = True
    if len(sys.argv) >= 3 and "--gunicorn" in sys.argv[2:]:
        uses_gunicorn = True
    if sys.argv[1] == "server":
        runServer(profiling=is_profiling, uses_gunicorn=uses_gunicorn)
    elif sys.argv[1] == "worker":
        runWorker(profiling=is_profiling)
    elif sys.argv[1] == "client":
        runClient()
    else:
        print("Unrecognized command, please use {{server, worker, client}}")
else:
    print(f"Usage: {sys.argv[0]} {{server, worker, client}}")
    print("Optionally use --profiling for {server, worker}")
    sys.exit()




import json
from typing import Any, TYPE_CHECKING, Dict, Iterable, List, Optional, Set, Tuple, Union

from mcrit.queue.LocalQueue import Job


class JobCollection(object):

    def __init__(self, jobs:List[Job]):
        self._jobs = jobs

    @property
    def count(self):
        return len(self._jobs)

    def getJobs(self):
        return self._jobs

    def getJobsForSample(self, sample_id:int, method=None, matching_only=False, finished_only=False) -> bool:
        selected_jobs = []
        for job in self._jobs:
            if matching_only and not job.is_matching_job:
                continue
            if finished_only and not job.is_finished:
                continue
            if method is not None and not job.method == method:
                continue
            if job.has_sample_id(sample_id):
                selected_jobs.append(job)
        return selected_jobs
    
    def filterToSampleIds(self, sample_ids:List[int]):
        filtered_jobs = []
        for job in self._jobs:
            if job.sample_id in sample_ids:
                filtered_jobs.append(job)
        self._jobs = filtered_jobs

    def filterToMethods(self, methods:List[str]):
        filtered_jobs = []
        for job in self._jobs:
            if job.method in methods:
                filtered_jobs.append(job)
        self._jobs = filtered_jobs

    def getJobById(self, job_id):
        for job in self._jobs:
            if job.job_id == job_id:
                return job



import logging
from dataclasses import asdict, dataclass, field

from mcrit.config.ConfigInterface import ConfigInterface, default_field
from mcrit.queue.QueueFactory import QueueFactory


@dataclass
class QueueConfig(ConfigInterface):
    # QUEUE_METHOD:... = QueueFactory.QUEUE_METHOD_FAKE
    QUEUE_METHOD: ... = QueueFactory.QUEUE_METHOD_MONGODB
    QUEUE_SERVER: str = "127.0.0.1"
    QUEUE_PORT: str = "27017"
    # By default, MongoDbStorage's DB's name and MongoQueue's DB's name are both "mcrit"
    # Changing one DB name here or at runtime DOES NOT change the other name!
    QUEUE_MONGODB_DBNAME: str = "mcrit"
    QUEUE_MONGODB_USERNAME: str = None
    QUEUE_MONGODB_PASSWORD: str = None
    QUEUE_MONGODB_FLAGS: str = ""
    QUEUE_MONGODB_COLLECTION_NAME: str = "queue"
    QUEUE_TIMEOUT: int = 300
    QUEUE_MAX_ATTEMPTS: int = 3
    # QUEUE_CLEAN_INTERVAL is the time EACH WORKER waits between cleaning
    QUEUE_CLEAN_INTERVAL: int = 20 * 60  # Clean every 20 minutes



import datetime
from typing import TYPE_CHECKING, Dict, Optional


class FunctionLabelEntry(object):
    # MCRIT specific
    function_id: int
    function_label: str
    username: str
    timestamp: datetime

    def __init__(
        self,
        function_label: str,
        username: str,
        function_id: Optional[int]=None,
        timestamp: Optional[datetime.datetime]=None
    ) -> None:
        self.function_label = function_label
        self.username = username
        self.function_id = function_id
        self.timestamp = timestamp
        if timestamp is None:
            self.timestamp = datetime.datetime.utcnow()

    def setFunctionId(self, function_id):
        self.function_id = function_id

    def toDict(self):
        function_entry = {
            "function_label": self.function_label,
            "username": self.username,
            "timestamp": self.timestamp.strftime("%Y-%m-%dT%H:%M:%S"),
        }
        return function_entry

    @classmethod
    def fromDict(cls, entry_dict):
        function_label_entry = cls(None, None)  # type: ignore
        function_label_entry.function_label = entry_dict["function_label"]
        # function_entry.function_id = entry_dict["function_id"]
        function_label_entry.username = entry_dict["username"]
        function_label_entry.function_id = entry_dict["function_id"] if "function_id" in entry_dict else None
        function_label_entry.timestamp = datetime.datetime.strptime(entry_dict["timestamp"], "%Y-%m-%dT%H:%M:%S")
        return function_label_entry

    def __str__(self):
        return f"FunctionLabel: '{self.function_label}' -- by user: {self.username}"



import logging
import re

import falcon

from mcrit.server.utils import timing, jsonify
from mcrit.index.MinHashIndex import MinHashIndex
from mcrit.server.utils import db_log_msg


class FunctionResource:
    def __init__(self, index: MinHashIndex):
        self.index = index

    @timing
    def on_get(self, req, resp, function_id=None):
        query_with_xcfg = False 
        if "with_xcfg" in req.params:
            query_with_xcfg = req.params["with_xcfg"].lower().strip() == "true"
        if not self.index.isFunctionId(function_id):
            resp.data = jsonify(
                {
                    "status": "failed",
                    "data": {"message": "We don't have a function with that id."},
                }
            )
            resp.status = falcon.HTTP_404
            db_log_msg(self.index, req, f"FunctionResource.on_get - failed - {function_id} unknown.")
            return
        data = self.index.getFunctionById(function_id, with_xcfg=query_with_xcfg).toDict()
        resp.data = jsonify(
            {
                "status": "successful",
                "data": data,
            }
        )
        db_log_msg(self.index, req, f"FunctionResource.on_get - success - {function_id}.")

    @timing
    def on_post_collection(self, req, resp):
        if not req.content_length:
            resp.data = jsonify(
                {
                    "status": "failed",
                    "data": {"message": "POST request without body can't be processed."},
                }
            )
            resp.status = falcon.HTTP_400
            db_log_msg(self.index, req, f"FunctionResource.on_post - failed - no POST body.")
            return
        with_label_only = False 
        if "with_label_only" in req.params:
            with_label_only = req.params["with_label_only"].lower().strip() == "true"
        # assume the POST body consists of comma separated function_ids
        post_body = req.stream.read()
        if re.match(b"^\d+(?:[\s]*,[\s]*\d+)*$", post_body):
            target_function_ids = [int(function_id) for function_id in post_body.split(b",")]
            function_entries = {}
            for function_id in target_function_ids:
                function_entry = self.index.getFunctionById(function_id, with_xcfg=False).toDict()
                if function_entry:
                    if with_label_only and not function_entry["function_labels"]:
                        continue
                    function_entries[function_id] = function_entry
            resp.data = jsonify({"status": "successful", "data": function_entries})
            resp.status = falcon.HTTP_200
            db_log_msg(self.index, req, f"FunctionResource.on_post - success.")
            return
        resp.status = falcon.HTTP_400
        db_log_msg(self.index, req, f"FunctionResource.on_post - failed - invalid body format.")

    @timing
    def on_get_collection(self, req, resp):
        # parse optional request parameters
        start_index = 0 
        if "start" in req.params:
            try:
                start_index = int(req.params["start"])
            except:
                pass
        limit_function_count = 0 
        if "limit" in req.params:
            try:
                limit_function_count = int(req.params["limit"])
            except:
                pass
        function_overview = {}
        function_entries = self.index.getFunctions(start_index, limit_function_count)
        for function_entry in function_entries:
            function_overview[function_entry.function_id] = function_entry.toDict()
        resp.data = jsonify({"status": "successful", "data": function_overview})
        db_log_msg(self.index, req, f"FunctionResource.on_get_collection - success.")



from typing import TYPE_CHECKING, Dict, Optional

from smda.common.BinaryInfo import BinaryInfo
from smda.common.SmdaFunction import SmdaFunction

from mcrit.libs.utility import encode_two_complement, decode_two_complement
from mcrit.minhash.MinHash import MinHash
from mcrit.storage.FunctionLabelEntry import FunctionLabelEntry

if TYPE_CHECKING:  # pragma: no cover
    from mcrit.storage.SampleEntry import SampleEntry
    from smda.common.SmdaFunction import SmdaFunction

# Dataclass, post init
# constructor -> .fromSmdaFunction
# assume sample_entry, smda_function always available

# TODO remove (everywhere) from alchemy function


class FunctionEntry(object):
    # MCRIT specific
    function_id: int
    family_id: int
    sample_id: int
    minhash: bytes  # TODO rename -> minhash_bytes? minhash_hex?
    minhash_shingle_composition: Dict = None  # FIXME MongoDbStorage fails without this, ... why?
    # inherited from sample
    architecture: str
    # smda information
    function_name: str
    function_labels: list
    matches: Dict
    pichash: int
    picblockhashes: list
    num_blocks: int
    num_instructions: int
    binweight: float
    offset: int
    xcfg: Dict

    def __init__(
        self,
        sample_entry: "SampleEntry",
        smda_function: "SmdaFunction",
        function_id: int,
        minhash: Optional[MinHash] = None,
    ) -> None:
        self.function_id = function_id
        if sample_entry:
            self.family_id = sample_entry.family_id
            self.sample_id = sample_entry.sample_id
            self.architecture = sample_entry.architecture
        if smda_function:
            self.num_blocks = smda_function.num_blocks
            self.num_instructions = smda_function.num_instructions
            self.binweight = smda_function.binweight
            self.offset = smda_function.offset
            self.xcfg = smda_function.toDict()
            self.function_name = smda_function.function_name
            self.pichash = smda_function.pic_hash
            self.picblockhashes = []
        self.function_labels = []
        self.matches = {}
        empty_minhash = MinHash()
        self.minhash = minhash.getMinHash() if minhash else empty_minhash.getMinHash()
        self.shingler_composition = minhash.getComposition() if minhash else empty_minhash.getComposition()

    def getMinHash(self, minhash_bits=32):
        return MinHash(function_id=self.function_id, minhash_bytes=self.minhash, minhash_bits=minhash_bits)

    def toSmdaFunction(self):
        binary_info = BinaryInfo(b"")
        binary_info.architecture = self.architecture
        return SmdaFunction.fromDict(self.xcfg, binary_info=binary_info)

    def toDict(self):
        empty_minhash = MinHash()
        minhash = self.minhash if self.minhash else empty_minhash.getMinHash()
        shingler_composition = (
            self.minhash_shingle_composition if self.minhash_shingle_composition else empty_minhash.getComposition()
        )
        function_entry = {
            "architecture": self.architecture,
            "binweight": self.binweight,
            "family_id": self.family_id,
            "function_id": self.function_id,
            "function_name": self.function_name,
            "function_labels": [l.toDict() for l in self.function_labels],
            "matches": self.matches,
            "minhash": minhash.hex(),
            "minhash_shingle_composition": shingler_composition,
            "num_blocks": self.num_blocks,
            "num_instructions": self.num_instructions,
            "offset": encode_two_complement(self.offset),
            "pichash": self.pichash,
            "picblockhashes": self.picblockhashes,
            "sample_id": self.sample_id,
            "xcfg": self.xcfg,
        }
        return function_entry

    @classmethod
    def fromDict(cls, entry_dict):
        function_entry = cls(None, None, entry_dict["function_id"])  # type: ignore
        function_entry.family_id = entry_dict["family_id"]
        # function_entry.function_id = entry_dict["function_id"]
        function_entry.sample_id = entry_dict["sample_id"]
        function_entry.architecture = entry_dict["architecture"]
        function_entry.function_name = entry_dict["function_name"]
        function_entry.function_labels = [FunctionLabelEntry.fromDict(l) for l in entry_dict["function_labels"]] if "function_labels" in entry_dict else []
        for label in function_entry.function_labels:
            label.setFunctionId(entry_dict["function_id"])
        function_entry.matches = entry_dict["matches"]
        function_entry.pichash = entry_dict["pichash"]
        function_entry.picblockhashes = entry_dict["picblockhashes"]
        function_entry.minhash = bytes.fromhex(entry_dict["minhash"])
        function_entry.minhash_shingle_composition = entry_dict["minhash_shingle_composition"]
        function_entry.num_blocks = entry_dict["num_blocks"]
        function_entry.num_instructions = entry_dict["num_instructions"]
        function_entry.binweight = entry_dict["binweight"]
        function_entry.offset = decode_two_complement(entry_dict["offset"])
        function_entry.xcfg = entry_dict["xcfg"] if "xcfg" in entry_dict else None
        return function_entry

    @classmethod
    def fromAlchemyFunction(cls, function):
        function_entry = cls(None, None)  # type:ignore
        function_entry.binweight = function.binweight
        function_entry.function_id = function.id
        function_entry.family_id = function.family.id
        function_entry.sample_id = function.sample.id
        function_entry.architecture = function.architecture.name
        function_entry.function_name = function.function_name
        if function.pichash:
            function_entry.pichash = function.pichash.pichash
        if function.minhash:
            function_entry.minhash = function.minhash.minhash
        function_entry.num_blocks = function.num_blocks
        function_entry.num_instructions = function.num_instructions
        function_entry.offset = function.offset
        function_entry.xcfg = function.xcfg
        return function_entry

    def __str__(self):
        return "Family: {} Sample: {} Function: {} @ 0x{:08x} - {} blocks ({} hashes), {} instructions - pichash: {}".format(
            self.family_id,
            self.sample_id,
            self.function_id,
            self.offset,
            self.num_blocks,
            len(self.picblockhashes),
            self.num_instructions,
            self.pichash,
        )



import datetime
import functools
import logging
import math
from collections import defaultdict, Counter
from multiprocessing import Pool, cpu_count
from timeit import default_timer as timer
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Optional, Set, Tuple, Union

import tqdm
from mcrit.queue.QueueRemoteCalls import NoProgressReporter

if TYPE_CHECKING:  # pragma: no cover
    from mcrit.storage.FunctionEntry import FunctionEntry
    from mcrit.storage.SampleEntry import SampleEntry
    from mcrit.storage.MatchingCache import MatchingCache
    from mcrit.storage.MemoryStorage import MemoryStorage
    from mcrit.Worker import Worker


# Only do basicConfig if no handlers have been configured
if len(logging._handlerList) == 0:
    logging.basicConfig(level=logging.INFO, format="%(asctime)-15s %(message)s")
LOGGER = logging.getLogger(__name__)

# function_id_own, sample_id_foreign, function_id_foreign, score, is_pic_match, is_min_match
HarmonizedMatches = Dict[Tuple[int, int, int], Tuple[float, bool, bool]]
PichashMatches = Dict[int, Set[Tuple[int, int, int]]]
MinhashMatches = List[Tuple[int, int, int, int, float]]

IS_MINHASH_FLAG = 1
IS_PICHASH_FLAG = 1 << 1
IS_LIBRARY_FLAG = 1 << 2


def build_method_str_from_args(args):
    # extract the job parameters
    method_str = ""
    try:
        method_type = f"{type(args[0])}"[8:-2]
        method_str = method_type.split(".")[-1]
        if len(args) > 1:
            method_str += "(" + ", ".join([str(i) for i in list(args)[1:]]) + ")"
    except:
        pass
    return method_str


def add_duration(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        start = timer()
        result = func(*args, **kwargs)
        time = timer() - start
        result["info"]["job"] = {
            "duration": time, 
            "timestamp": datetime.datetime.now().isoformat(),
            "parameters": build_method_str_from_args(args)
        }
        return result

    return wrapper


class MatcherInterface(object):
    def __init__(
        self, worker: "Worker", 
        minhash_threshold=None, 
        pichash_size=None, 
        band_matches_required=None, 
        exclude_self_matches=False,
        progress_reporter=NoProgressReporter()
    ):
        # Extended by Query, VS, Sample
        self._worker: "Worker" = worker
        self._storage = worker.getStorage()
        self._function_entries: List["FunctionEntry"] = []
        self._sample_info: Optional[Dict] = None
        self._sample_id: Optional[int] = None
        self._progress_reporter = progress_reporter
        self._num_batches = None
        self._minhash_threshold = minhash_threshold
        self._exclude_self_matches = exclude_self_matches
        if pichash_size is None:
            pichash_size = self._worker.config.MINHASH_CONFIG.PICHASH_SIZE
        if band_matches_required is None:
            band_matches_required = self._worker.config.MINHASH_CONFIG.BAND_MATCHES_REQUIRED
        self._band_matches_required = band_matches_required
        self._pichash_size = pichash_size
        self._additional_setup()
        self._sample_to_lib_info: Dict[int, bool]
        self._sample_id_to_entry: Dict[int, SampleEntry]

    def _additional_setup(self):
        pass

    def getMatchesForSample(self, sample_id, other_sample_id, minhash_threshold=None):
        # VS, Sample
        raise NotImplementedError

    def getMatchesForSmdaReport(self, smda_report, minhash_threshold=None):
        # Query
        raise NotImplementedError

    def _getPicHashMatches(self) -> Dict[int, Set[Tuple[int, int]]]:
        # Query, VS, Sample
        # Every Matcher has own implementation
        raise NotImplementedError

    def _createMinHashCandidateGroups(self, start=0, end=None) -> Dict[int, Set[int]]:
        # Query, VS, Sample
        # Sample , Query have this plain version
        # VS adds another intersection step to this version
        # find candidates based on bands
        if end is None:
            end = len(self._function_entries)
        candidate_groups = {}
        function_id_to_minhash = {}
        for function_entry in self._function_entries[start:end]:
            function_id_to_minhash[function_entry.function_id] = function_entry.getMinHash(
                minhash_bits=self._worker._minhash_config.MINHASH_SIGNATURE_BITS
            )
        candidate_groups = self._storage.getCandidatesForMinHashes(function_id_to_minhash, band_matches_required=self._band_matches_required)

        return candidate_groups

    def _createMatchingCache(self, candidate_groups):
        # Query, VS, Sample
        # This version is used by VS, Sample
        # Own version by Query
        cache_function_ids = set()
        for candidate_key, candidate_functions in candidate_groups.items():
            cache_function_ids.add(candidate_key)
            cache_function_ids.update(candidate_functions)
        LOGGER.info("creating cache for %d functions", len(cache_function_ids))
        return self._storage.createMatchingCache(cache_function_ids)

    ######## Below this line, nothing will be overwritten by Subclasses #########

    def filter_pichashes_from_candidate_groups(self, matching_cache, candidate_groups, pichash_matches):
        finished_tuples = set()
        for pichash_match in pichash_matches:
            own_fid, foreign_sid, foreign_fid = pichash_match
            if own_fid in candidate_groups:
                # remove all candidates of the respective sample instead
                if (own_fid, foreign_sid) not in finished_tuples:
                    if matching_cache.isSampleId(foreign_sid):
                        pichash_match_sid_fids = matching_cache.getFunctionIdsBySampleId(foreign_sid)
                        candidate_groups[own_fid].difference_update(pichash_match_sid_fids)
                        finished_tuples.add((own_fid, foreign_sid))
        return candidate_groups

    def _getMatchesRoutine(self):
        # Query, VS, Sample
        # All use this version
        pichash_matches = self._harmonizePicHashMatches(self._getPicHashMatches())
        LOGGER.info("Calculated PicHash matches")
        all_minhash_matches = {}
        # only do minhashing, if we have bands to evaluate
        if self._band_matches_required > 0:
            # if we have an exceedingly large number of functions, we need to process in batches...
            quotient, remainder = divmod(len(self._function_entries), self._worker.config.MINHASH_CONFIG.MINHASH_MATCHING_FUNCTION_BATCH_SIZE)
            self._num_batches = quotient + int(bool(remainder)) # always round up
            self._progress_reporter.set_total(self._num_batches)
            for start_index in range(0, len(self._function_entries), self._worker.config.MINHASH_CONFIG.MINHASH_MATCHING_FUNCTION_BATCH_SIZE):
                candidate_groups = self._createMinHashCandidateGroups(start=start_index, end=start_index+self._worker.config.MINHASH_CONFIG.MINHASH_MATCHING_FUNCTION_BATCH_SIZE)
                LOGGER.info("Created candidate groups from MinHash bands")
                matching_cache = self._createMatchingCache(candidate_groups)
                LOGGER.info("Created MatchingCache")
                if self._worker._minhash_config.PICHASH_IMPLIES_MINHASH_MATCH:
                    LOGGER.info("Removing PicHash matches")
                    candidate_groups = self.filter_pichashes_from_candidate_groups(matching_cache, candidate_groups, pichash_matches)
                    LOGGER.info("Removed PicHash matches from CandidateGroups")
                LOGGER.info("Now starting MinHash matching")
                minhash_matches = self._harmonizeMinHashMatches(self._sample_id, self._performMinHashMatching(candidate_groups, matching_cache))
                all_minhash_matches.update(minhash_matches)
                if self._num_batches > 1:
                    self._progress_reporter.step()
            LOGGER.info("Calculated MinHash matches.")
        matching_report = self._craftResultDict(pichash_matches, all_minhash_matches)
        LOGGER.info("Returning aggregated match report.")
        return matching_report

    # Reports PROGRESS
    def _performMinHashMatching(self, candidate_groups: Dict[int, Set[int]], cache) -> List[Tuple[int, int, int, int, float]]:
        # Query, VS, Sample
        # All use the same version
        """perform matching between candidates, provided as candidate_groups in the form of a dict: {function_id: [function_ids]}"""
        # result format: sample_id_a, function_id_a, sample_id_b, function_id_b, score
        matching_results: List[Tuple[int, int, int, int, float]] = []
        organized_matching_results: Dict[Tuple[int, int, int], Tuple[int, float]] = defaultdict(lambda: (-1, -1.0))
        packed_tuples = self._unrollGroupsAsPackedTuples(cache, candidate_groups)
        num_packed_tuples = self._countPackedTuples(candidate_groups)
        # progress reporter counts here in case calling function does not count
        single_batch = False
        if self._num_batches == 1:
            self._progress_reporter.set_total(num_packed_tuples)
            single_batch = True
        calculation_function = functools.partial(
            self._worker.minhasher.calculateScoresFromPackedTuples, ignore_threshold=True, minhash_threshold=self._minhash_threshold
        )
        packed_tuples = [p for p in packed_tuples]
        counted_scores = Counter()
        if self._worker._minhash_config.MINHASH_POOL_MATCHING:
            with Pool(cpu_count()) as pool:
                for pool_result in tqdm.tqdm(
                    pool.imap_unordered(calculation_function, packed_tuples),
                    total=num_packed_tuples,
                ):
                    # for key, new_value in pool_result.items():
                    for single_result in pool_result:
                        sample_id_a, function_id_a, sample_id_b, function_id_b, score = single_result
                        counted_scores[score] += 1
                        if score > self._worker.config.MINHASH_CONFIG.MINHASH_MATCHING_THRESHOLD:
                            key = (sample_id_a, function_id_a, sample_id_b)
                            new_value = (function_id_b, score)
                            original_value = organized_matching_results[key]
                            organized_matching_results[key] = max([original_value, new_value], key=lambda x:x[1])
                    if single_batch:
                        self._progress_reporter.step()
        else:
            packed_tuple: List[Tuple[int, int, bytes, int, int, bytes]]
            for packed_tuple in tqdm.tqdm(packed_tuples, total=num_packed_tuples):
                pool_result = calculation_function(packed_tuple)
                for single_result in pool_result:
                    sample_id_a, function_id_a, sample_id_b, function_id_b, score = single_result
                    counted_scores[score] += 1
                    if score > self._worker.config.MINHASH_CONFIG.MINHASH_MATCHING_THRESHOLD:
                        key = (sample_id_a, function_id_a, sample_id_b)
                        new_value = (function_id_b, score)
                        original_value = organized_matching_results[key]
                        organized_matching_results[key] = max([original_value, new_value], key=lambda x: x[1])
                if single_batch:
                    self._progress_reporter.step()
        full_score_counts = sorted([(item[0]*64/100, item[1]) for item in dict(counted_scores).items()])
        LOGGER.info("Minhash Signature Field Match Counts: " + ", ".join([f"({i[0]}: {i[1]})" for i in full_score_counts]))
        matching_results = [k+v for k, v in organized_matching_results.items()]
        self._storage.clearMatchingCache()
        return matching_results

    def _countPackedTuples(self, candidate_pairs) -> int:
        count = 0
        for candidate_ids in candidate_pairs.values():
            count += len(candidate_ids) 
        quotient, remainder = divmod(count, self._worker.config.MINHASH_CONFIG.MINHASH_MATCHING_CANDIDATE_WORKPACK_SIZE)
        LOGGER.info("Processing a total of %d candidates.", count)
        return quotient + int(bool(remainder)) # always round up

    def _unrollGroupsAsPackedTuples(
        self, cache: Union["MatchingCache", "MemoryStorage"], candidate_pairs
    ) -> Iterable[List[Tuple[int, int, bytes, int, int, bytes]]]:
        # Query, VS, Sample
        # All were identical
        packed_tuples: List[List[Tuple[int, int, bytes, int, int, bytes]]] = []
        current_pack: List[Tuple[int, int, bytes, int, int, bytes]] = []
        count = 0
        for function_id_a, candidate_ids in sorted(candidate_pairs.items()):
            for function_id_b in sorted(candidate_ids):
                if function_id_a == function_id_b:
                    continue
                minhash_a = cache.getMinHashByFunctionId(function_id_a)
                minhash_b = cache.getMinHashByFunctionId(function_id_b)
                sample_id_a = cache.getSampleIdByFunctionId(function_id_a)
                sample_id_b = cache.getSampleIdByFunctionId(function_id_b)
                current_pack.append((sample_id_a, function_id_a, minhash_a, sample_id_b, function_id_b, minhash_b))
                if count < self._worker.config.MINHASH_CONFIG.MINHASH_MATCHING_CANDIDATE_WORKPACK_SIZE:
                    count += 1
                else:
                    yield current_pack
                    current_pack = []
                    count = 0
        if current_pack:
            yield current_pack
        return packed_tuples

    def _harmonizePicHashMatches(self, pichash_matches: PichashMatches) -> HarmonizedMatches:
        # Query, VS, Sample
        # all use this version
        pichash_match_mapping: HarmonizedMatches = {}
        cached_function_id_to_sample_id: Dict[int, int] = {}
        # all_own_function_ids = {entry.function_id for entry in self._function_entries}
        sample_function_id_to_size = {entry.function_id: entry.num_instructions for entry in self._function_entries}
        for _, pichash_tuples in pichash_matches.items():
            current_own_function_ids: List[int] = []
            current_foreign_function_ids: List[int] = []
            function_size = 0
            for pichash_tuple in pichash_tuples:
                bh_family_id, bh_sample_id, bh_function_id = pichash_tuple
                cached_function_id_to_sample_id[bh_function_id] = bh_sample_id
                if bh_function_id in sample_function_id_to_size:
                    current_own_function_ids.append(bh_function_id)
                    function_size = sample_function_id_to_size[bh_function_id]
                else:
                    current_foreign_function_ids.append(bh_function_id)
            if function_size < self._pichash_size:
                continue
            # eval matches for this set of tuples
            if len(pichash_tuples) > 1:
                for own_function_id in current_own_function_ids:
                    for foreign_function_id in current_foreign_function_ids:
                        foreign_sample_id = cached_function_id_to_sample_id[foreign_function_id]
                        foreign_tuple = (foreign_sample_id, foreign_function_id)
                        pichash_match_mapping[(own_function_id, *foreign_tuple)] = (100, True, False)
                    # include self matches
                    for own_function_id_2 in current_own_function_ids:
                        if own_function_id == own_function_id_2:
                            continue
                        foreign_sample_id = cached_function_id_to_sample_id[own_function_id_2]
                        foreign_tuple = (foreign_sample_id, own_function_id_2)
                        pichash_match_mapping[(own_function_id, *foreign_tuple)] = (100, True, False)
        return pichash_match_mapping

    def _harmonizeMinHashMatches(self, sample_id, minhash_matches: MinhashMatches) -> HarmonizedMatches:
        # Query, VS, Sample
        # all use this version
        # handle minhashes
        minhash_mapping: HarmonizedMatches = {}

        for match in minhash_matches:
            sample_id_a, function_id_a, sample_id_b, function_id_b, minhash_score = match
            if sample_id_a == sample_id:
                own_function_id = function_id_a
                foreign_function_id = function_id_b
                foreign_sample_id = sample_id_b
                minhash_mapping[(own_function_id, foreign_sample_id, foreign_function_id)] = (
                    minhash_score,
                    False,
                    True,
                )
            if sample_id_b == sample_id:
                own_function_id = function_id_b
                foreign_function_id = function_id_a
                foreign_sample_id = sample_id_a
                minhash_mapping[(own_function_id, foreign_sample_id, foreign_function_id)] = (
                    minhash_score,
                    False,
                    True,
                )

        return minhash_mapping

    # summarizing, formatting starts here:

    def _summarizeMatches(
        self, sample_id, matches: HarmonizedMatches, aggregation_only: bool
    ) -> Tuple[List, Dict, Dict, float]:
        # Query, VS, Sample
        # all use this version
        sample_fid_to_binweight = {entry.function_id: entry.binweight for entry in self._function_entries}
        sample_fid_to_offset = {entry.function_id: entry.offset for entry in self._function_entries}
        aggregation = {
            "num_own_functions_matched": 0,
            "num_foreign_functions_matched": 0,
            "num_own_functions_matched_as_library": 0,
            "num_self_matches": 0,
            "bytes_matched": 0.0,
        }
        # handle minhashes
        match_function_mapping: Dict[int, Dict] = {}
        match_sample_mapping: Dict[int, Any]
        matches_sample_list = []
        # from here on structure the stuff in the same way as pichash result
        self_matched_functions: Set[int] = set()
        own_functions_with_matches: Set[int] = set()
        own_functions_with_library_matches: Set[int] = set()
        foreign_functions_matched: Set[int] = set()

        for match_ids, match_strength in matches.items():
            own_function_id, foreign_sample_id, foreign_function_id = match_ids
            minhash_score, is_pichash_match, is_minhash_match = match_strength

            if foreign_sample_id == sample_id:
                self_matched_functions.add(own_function_id)
                self_matched_functions.add(foreign_function_id)
            else:
                if foreign_sample_id not in self._sample_to_lib_info:
                    self._sample_to_lib_info[foreign_sample_id] = (
                        self._storage.getLibraryInfoForSampleId(foreign_sample_id) is not None
                    )
                has_libinfo = self._sample_to_lib_info[foreign_sample_id]

                if foreign_sample_id not in self._sample_id_to_entry:
                    self._sample_id_to_entry[foreign_sample_id] = self._storage.getSampleById(foreign_sample_id)
                foreign_family_id = self._sample_id_to_entry[foreign_sample_id].family_id

                if own_function_id not in match_function_mapping:
                    match_function_mapping[own_function_id] = {
                        "num_bytes": sample_fid_to_binweight[own_function_id],
                        "offset": sample_fid_to_offset[own_function_id],
                        "matches": [],
                    }

                flags = (
                    is_pichash_match * IS_PICHASH_FLAG
                    + is_minhash_match * IS_MINHASH_FLAG
                    + has_libinfo * IS_LIBRARY_FLAG
                )
                match_function_mapping[own_function_id]["matches"].append(
                    (
                        foreign_family_id,
                        foreign_sample_id,
                        foreign_function_id,
                        minhash_score,
                        flags,
                    )
                )
                foreign_functions_matched.add(foreign_function_id)
                own_functions_with_matches.add(own_function_id)
                if has_libinfo:
                    own_functions_with_library_matches.add(own_function_id)

        # Anti-Value-Key post processing
        matches_function_list = []
        for function_id, dict in match_function_mapping.items():
            dict["fid"] = function_id
            # this creates additional stability for tests and processing
            dict["matches"] = sorted(dict["matches"])
            matches_function_list.append(dict)

        aggregation["num_self_matches"] = len(self_matched_functions)
        aggregation["num_own_functions_matched"] = len(own_functions_with_matches)
        aggregation["bytes_matched"] = sum(
            [sample_fid_to_binweight[own_function_id] for own_function_id in own_functions_with_matches]
        )
        aggregation["num_foreign_functions_matched"] = len(foreign_functions_matched)
        aggregation["num_own_functions_matched_as_library"] = len(own_functions_with_library_matches)

        num_library_match_bytes = sum(
            [match_function_mapping[function_id]["num_bytes"] for function_id in own_functions_with_library_matches]
        )

        return matches_function_list, match_function_mapping, aggregation, num_library_match_bytes

    def _get_family_adjustment(self, match_report) -> Dict[int, float]:
        adjustments: Dict[int, float] = {}
        for fid, function_data in match_report.items():
            families = {match[0] for match in function_data["matches"]}
            family_adjustment_value = 1 if len(families) < 3 else 1 + int(math.log(len(families), 2))
            adjustments[fid] = family_adjustment_value
        return adjustments

    def _aggregateMatchSampleSummary(self, match_report, own_sample_info, num_library_bytes):
        own_sample_num_bytes = own_sample_info["binweight"]
        own_sample_num_nonlibrary_bytes = own_sample_num_bytes - num_library_bytes
        # aggregate sample byte sizes
        function_num_bytes = {}
        for own_function_id, function_data in match_report.items():
            function_num_bytes[own_function_id] = function_data["num_bytes"]
        # summarize samples
        matches_per_sample = self._aggregateMatchesPerSample(match_report)
        sample_summary = {}

        family_adjustment = self._get_family_adjustment(match_report)

        for foreign_sample_id in matches_per_sample:
            if not foreign_sample_id in self._sample_id_to_entry:
                self._sample_id_to_entry[foreign_sample_id] = self._storage.getSampleById(foreign_sample_id)
            sample_info = self._sample_id_to_entry[foreign_sample_id]
            sample_summary[foreign_sample_id] = {
                "family": sample_info.family,
                "family_id": sample_info.family_id,
                "version": sample_info.version,
                "bitness": sample_info.bitness,
                "sha256": sample_info.sha256,
                "filename": sample_info.filename,
                "sample_id": foreign_sample_id,
                "num_bytes": sample_info.binweight,
                "is_library": sample_info.is_library,
                "num_functions": sample_info.statistics["num_functions"],
                "matched": {
                    "functions": {
                        "minhashes": 0,
                        "pichashes": 0,
                        "combined": 0,
                        "library": 0,
                    },
                    "bytes": {
                        "unweighted": 0,
                        "score_weighted": 0,
                        "frequency_weighted": 0,
                        "nonlib_unweighted": 0,
                        "nonlib_score_weighted": 0,
                        "nonlib_frequency_weighted": 0,
                    },
                    "percent": {
                        "unweighted": 0,
                        "score_weighted": 0,
                        "frequency_weighted": 0,
                        "nonlib_unweighted": 0,
                        "nonlib_score_weighted": 0,
                        "nonlib_frequency_weighted": 0,
                    },
                },
            }
            current_matched = sample_summary[foreign_sample_id]["matched"]
            for own_function_id, matches in matches_per_sample[foreign_sample_id].items():
                current_matched["functions"]["combined"] += 1
                has_library_match = "library" in [match[0] for match in matches]
                current_matched["functions"]["library"] += 1 if has_library_match else 0
                current_matched["functions"]["minhashes"] += 1 if "minhash" in [match[0] for match in matches] else 0
                current_matched["functions"]["pichashes"] += 1 if "pichash" in [match[0] for match in matches] else 0

                match_max = max([match[1] for match in matches])
                unweighted_inc = function_num_bytes[own_function_id]
                score_weighted_inc = 1.0 * unweighted_inc * match_max / 100.0
                frequency_weighted_inc = score_weighted_inc / family_adjustment[own_function_id]

                current_matched["bytes"]["unweighted"] += unweighted_inc
                current_matched["bytes"]["score_weighted"] += score_weighted_inc
                current_matched["bytes"]["frequency_weighted"] += frequency_weighted_inc
                if not has_library_match:
                    current_matched["bytes"]["nonlib_unweighted"] += unweighted_inc
                    current_matched["bytes"]["nonlib_score_weighted"] += score_weighted_inc
                    current_matched["bytes"]["nonlib_frequency_weighted"] += frequency_weighted_inc

            for kind in "unweighted", "score_weighted", "frequency_weighted":
                current_matched["percent"][kind] = 100.0 * current_matched["bytes"][kind] / own_sample_num_bytes
                current_matched["percent"]["nonlib_" + kind] = (
                    100.0 * current_matched["bytes"]["nonlib_" + kind] / own_sample_num_nonlibrary_bytes
                )

        return list(sample_summary.values())

    def _aggregateMatchesPerSample(self, match_report):
        matches_per_sample: Dict[int, Dict[int, List[Tuple[str, float]]]] = defaultdict(lambda: defaultdict(list))
        for own_function_id, function_data in match_report.items():
            function_has_libinfo = any([bool(match[-1] & IS_LIBRARY_FLAG) for match in function_data["matches"]])
            for match in function_data["matches"]:
                (
                    foreign_family_id,
                    foreign_sample_id,
                    foreign_function_id,
                    minhash_score,
                    flags,
                ) = match
                is_pichash_match = flags & IS_PICHASH_FLAG
                is_minhash_match = flags & IS_MINHASH_FLAG
                match_types = []
                if is_pichash_match:
                    match_types.append("pichash")
                if is_minhash_match:
                    match_types.append("minhash")
                for match_type in match_types:
                    matches_per_sample[foreign_sample_id][own_function_id].append((match_type, minhash_score))
                if function_has_libinfo:
                    matches_per_sample[foreign_sample_id][own_function_id].append(("library", 0))
        return matches_per_sample

    def _craftResultDict(
        self, pichash_matches: HarmonizedMatches, minhash_matches: HarmonizedMatches, num_matches=None
    ) -> Dict:
        # Query, VS, Sample
        # All use this version
        if self._worker._minhash_config.PICHASH_IMPLIES_MINHASH_MATCH:
            for key, new_entry in pichash_matches.items():
                if key not in minhash_matches:
                    minhash_matches[key] = (100.0, 0, 1)
        all_matches = minhash_matches.copy()
        for key, new_entry in pichash_matches.items():
            if key in all_matches:
                old_entry = all_matches[key]
                # update values: score, is_pic, is_min
                all_matches[key] = tuple([max(old_entry[i], new_entry[i]) for i in range(3)])
            else:
                all_matches[key] = new_entry

        _, _, pichash_aggregation, _ = self._summarizeMatches(self._sample_id, pichash_matches, False)
        # Assume that every pichash is also a min hash:
        _, _, minhash_aggregation, _ = self._summarizeMatches(self._sample_id, minhash_matches, False)

        (all_functions_summary, all_functions_list, all_aggregation, num_library_bytes) = self._summarizeMatches(
            self._sample_id, all_matches, False
        )
        sample_summary = self._aggregateMatchSampleSummary(all_functions_list, self._sample_info, num_library_bytes)
        summary = {
            "info": {
                "job": None,
                "sample": self._sample_info,
            },
            "matches": {
                "aggregation": {
                    # assume all pichashes are also minhashes
                    "minhash": minhash_aggregation,
                    "pichash": pichash_aggregation,
                },
                "functions": all_functions_summary,
                "samples": sample_summary,
            },
        }
        return summary



# -*- coding: utf-8 -*-

from setuptools import find_packages, setup

with open("README.md") as f:
    README = f.read()

setup(
    name='mcrit',
    version="1.3.11",
    description='MCRIT is a framework created for simplified application of the MinHash algorithm to code similarity.',
    long_description_content_type="text/markdown",
    long_description=README,
    author='Daniel Plohmann, Manuel Blatt, Steffen Enders, Paul Hordiienko',
    author_email='daniel.plohmann@fkie.fraunhofer.de',
    url='https://github.com/danielplohmann/mcrit',
    license="NU General Public License v3 (GPLv3)",
    packages=find_packages(exclude=("tests", "data", "docs", "examples", "plugins")),
    install_requires=open("requirements.txt").read().splitlines(),
    data_files=[
        ("", ["LICENSE", "requirements.txt"]),
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Security",
        "Topic :: Scientific/Engineering :: Information Analysis",
    ],
)



import struct
import hashlib
import logging 

from rapidfuzz.distance import Levenshtein
from smda.intel.IntelInstructionEscaper import IntelInstructionEscaper



class FunctionCfgMatcher(object):

    def __init__(self, sample_entry_a, smda_function_a, sample_entry_b, smda_function_b) -> None:
        """
        Initialize with two smda_functions and their respective sample_entries or smda_reports (only needed for base_addr and binary_size)
        """
        self.sample_entry_a = sample_entry_a
        self.sample_entry_b = sample_entry_b
        self.smda_function_a = smda_function_a
        self.smda_function_b = smda_function_b
        self.match_colors = {
            # based on hashing
            "regular_pic": "#00DDFF",
            "small_pic": "#C0F4FF",
            "escaped": "#00ff00",
            # based on distance function
            0: "#40ff40",
            1: "#c0ff80",
            2: "#FFFF40",
            3: "#FFCC40",
            99: "#FFA0A0",
        }

    @staticmethod
    def getPicBlockHashesForFunction(sample_entry, smda_function, min_size=0):
        pic_block_hashes = []
        for block in smda_function.getBlocks():
            if block.length >= min_size:
                escaped_binary_seq = []
                for instruction in block.getInstructions():
                    escaped_binary_seq.append(instruction.getEscapedBinary(IntelInstructionEscaper, escape_intraprocedural_jumps=True, lower_addr=sample_entry.base_addr, upper_addr=sample_entry.base_addr + sample_entry.binary_size))
                as_bytes = bytes([ord(c) for c in "".join(escaped_binary_seq)])
                hashed = struct.unpack("Q", hashlib.sha256(as_bytes).digest()[:8])[0]
                pic_block_hashes.append({"offset": block.offset, "hash": hashed, "size": block.length})
        return pic_block_hashes

    def getAllPicblockMatches(self):
        node_colors = {"a": {}, "b": {}}
        all_phbs_a = []
        for block in self.smda_function_a.getBlocks():
            escaped_binary_seq = []
            for instruction in block.getInstructions():
                escaped_binary_seq.append(instruction.getEscapedBinary(IntelInstructionEscaper, escape_intraprocedural_jumps=True, lower_addr=self.sample_entry_a.base_addr, upper_addr=self.sample_entry_a.base_addr + self.sample_entry_a.binary_size))
            as_bytes = bytes([ord(c) for c in "".join(escaped_binary_seq)])
            hashed = struct.unpack("Q", hashlib.sha256(as_bytes).digest()[:8])[0]
            all_phbs_a.append({"offset": block.offset, "hash": hashed, "size": block.length})
        all_phbs_b = []
        for block in self.smda_function_b.getBlocks():
            escaped_binary_seq = []
            for instruction in block.getInstructions():
                escaped_binary_seq.append(instruction.getEscapedBinary(IntelInstructionEscaper, escape_intraprocedural_jumps=True, lower_addr=self.sample_entry_b.base_addr, upper_addr=self.sample_entry_b.base_addr + self.sample_entry_b.binary_size))
            as_bytes = bytes([ord(c) for c in "".join(escaped_binary_seq)])
            hashed = struct.unpack("Q", hashlib.sha256(as_bytes).digest()[:8])[0]
            all_phbs_b.append({"offset": block.offset, "hash": hashed, "size": block.length})
        phb_a = set([pbh["hash"] for pbh in all_phbs_a])
        phb_b = set([pbh["hash"] for pbh in all_phbs_b])
        phb_match_addr_a = [(pbh["offset"], pbh["size"]) for pbh in all_phbs_a if pbh["hash"] in phb_a.intersection(phb_b)]
        phb_match_addr_b = [(pbh["offset"], pbh["size"]) for pbh in all_phbs_b if pbh["hash"] in phb_a.intersection(phb_b)]
        for addr, size in phb_match_addr_a:
            if size >= 4:
                node_colors["a"][f"Node0x{addr:x}"] = self.match_colors["regular_pic"]
            else:
                node_colors["a"][f"Node0x{addr:x}"] = self.match_colors["small_pic"]
        for addr, size in phb_match_addr_b:
            if size >= 4:
                node_colors["b"][f"Node0x{addr:x}"] = self.match_colors["regular_pic"]
            else:
                node_colors["b"][f"Node0x{addr:x}"] = self.match_colors["small_pic"]
        return node_colors

    def getEscapedMatches(self):
        node_colors = {"a": {}, "b": {}}
        all_escapes_a = []
        for block in self.smda_function_a.getBlocks():
            escaped_ins_seq = []
            for instruction in block.getInstructions():
                escaped_ins = IntelInstructionEscaper.escapeMnemonic(instruction.mnemonic) + " " + IntelInstructionEscaper.escapeOperands(instruction)
                escaped_ins_seq.append(escaped_ins)
            merged = ";".join(escaped_ins_seq)
            # print("0x%x" % block.offset, merged)
            hashed = struct.unpack("Q", hashlib.sha256(merged.encode("ascii")).digest()[:8])[0]
            all_escapes_a.append({"offset": block.offset, "hash": hashed})
        all_escapes_b = []
        for block in self.smda_function_b.getBlocks():
            escaped_ins_seq = []
            for instruction in block.getInstructions():
                escaped_ins = IntelInstructionEscaper.escapeMnemonic(instruction.mnemonic) + " " + IntelInstructionEscaper.escapeOperands(instruction)
                escaped_ins_seq.append(escaped_ins)
            merged = ";".join(escaped_ins_seq)
            # print("0x%x" % block.offset, merged)
            hashed = struct.unpack("Q", hashlib.sha256(merged.encode("ascii")).digest()[:8])[0]
            all_escapes_b.append({"offset": block.offset, "hash": hashed})
        phb_a = set([pbh["hash"] for pbh in all_escapes_a])
        phb_b = set([pbh["hash"] for pbh in all_escapes_b])
        phb_match_addr_a = [pbh["offset"] for pbh in all_escapes_a if pbh["hash"] in phb_a.intersection(phb_b)]
        phb_match_addr_b = [pbh["offset"] for pbh in all_escapes_b if pbh["hash"] in phb_a.intersection(phb_b)]
        for addr in phb_match_addr_a:
            node_colors["a"][f"Node0x{addr:x}"] = self.match_colors["escaped"]
        for addr in phb_match_addr_b:
            node_colors["b"][f"Node0x{addr:x}"] = self.match_colors["escaped"]
        return node_colors

    def getLevenshteinMatches(self, unmatched_nodes):
        node_colors = {"a": {}, "b": {}}
        # across all blocks in unmatched nodes, collect tokens and map to symbols
        # token -> symbol, like "M REG, REG" -> 0
        # we use symbols from chr(0x20) to chr(0x7e), i.e. up to 94 printables, which "should always be enough (TM)""
        alphabet = {}
        num_symbols = 0
        # offset -> symbolified block
        candidate_blocks_a = {}
        for block in self.smda_function_a.getBlocks():
            if block.offset not in unmatched_nodes["a"]:
                continue
            symbolified_block = ""
            for instruction in block.getInstructions():
                escaped_ins = instruction.mnemonic + " " + IntelInstructionEscaper.escapeOperands(instruction)
                if escaped_ins not in alphabet:
                    alphabet[escaped_ins] = chr(0x30 + num_symbols)
                    num_symbols += 1
                    if num_symbols > 94:
                        raise Exception("Basic Block contains too many tokens to compare.")
                symbolified_block += alphabet[escaped_ins]
            candidate_blocks_a[block.offset] = symbolified_block
        candidate_blocks_b = {}
        for block in self.smda_function_b.getBlocks():
            if block.offset not in unmatched_nodes["b"]:
                continue
            symbolified_block = ""
            for instruction in block.getInstructions():
                escaped_ins = instruction.mnemonic + " " + IntelInstructionEscaper.escapeOperands(instruction)
                if escaped_ins not in alphabet:
                    alphabet[escaped_ins] = chr(0x30 + num_symbols)
                    num_symbols += 1
                    if num_symbols > 94:
                        raise Exception("Basic Block contains too many tokens to compare.")
                symbolified_block += alphabet[escaped_ins]
            candidate_blocks_b[block.offset] = symbolified_block
        # print(alphabet)
        by_score = {0: [], 1: [], 2: [], 3: []}
        for block_a, symbols_a in candidate_blocks_a.items():
            for block_b, symbols_b in candidate_blocks_b.items():
                distance = Levenshtein.distance(symbols_a, symbols_b, score_cutoff=3)
                if distance < 4:
                    by_score[distance].append((block_a, block_b))
                    # print(f"0x{block_a:x} 0x{block_b:x}: {symbols_a} || {symbols_b} - {distance}")
        used_blocks = set()
        for score, pairs in by_score.items():
            for pair in pairs:
                block_a, block_b = pair
                if block_a not in used_blocks and block_b not in used_blocks:
                    node_colors["a"][f"Node0x{block_a:x}"] = score
                    node_colors["b"][f"Node0x{block_b:x}"] = score
                    used_blocks.add(block_a)
                    used_blocks.add(block_b)
        node_colors["a"] = {k: self.match_colors[v] for k, v in node_colors["a"].items()}
        node_colors["b"] = {k: self.match_colors[v] for k, v in node_colors["b"].items()}
        return node_colors

    def getColoredMatches(self):
        # thresholded edit distance match over escaped instruction sequence: green to orange
        node_colors = {"a": {}, "b": {}}
        # no match / base color: bleak red
        for block in self.smda_function_a.getBlocks():
            node_colors["a"][f"Node0x{block.offset:x}"] = self.match_colors[99]
        for block in self.smda_function_b.getBlocks():
            node_colors["b"][f"Node0x{block.offset:x}"] = self.match_colors[99]
        # escaped blocks matches
        escaped_block_matches = self.getEscapedMatches()
        node_colors["a"].update(escaped_block_matches["a"])
        node_colors["b"].update(escaped_block_matches["b"])
        # ad-hoc picblock match (small BB): bleak teal
        # full picblock match (regular BB>=4 instructions): teal
        smaller_picblock_matches = self.getAllPicblockMatches()
        node_colors["a"].update(smaller_picblock_matches["a"])
        node_colors["b"].update(smaller_picblock_matches["b"])
        # compare everything not colored by now using our adapted Levenshtein
        unmatched_nodes = {
            "a": [int(k[6:], 16) for k, v in node_colors["a"].items() if v == self.match_colors[99]], 
            "b": [int(k[6:], 16) for k, v in node_colors["b"].items() if v == self.match_colors[99]], 
        }
        levenshtein_matches = self.getLevenshteinMatches(unmatched_nodes)
        node_colors["a"].update(levenshtein_matches["a"])
        node_colors["b"].update(levenshtein_matches["b"])
        return node_colors
